#include "scene.h"
#include "triangle.h"
#include "tmesh.h"
#include <float.h>

Scene *scene;

#include <iostream>

using namespace std;


Scene::Scene() {


  markersN = 4;
  markers = new V3[markersN];
  markers[0] = V3(82.0f, 174.0f, 0.0f);
  markers[1] = V3(281.0f, 152.0f, 0.0f);
  markers[2] = V3(628.0f, 265.0f, 0.0f);
  markers[3] = V3(293.0f, 374.0f, 0.0f);
    
  photo = 0;

  tm1 = tm2 = 0;
  gui = new GUI();
  gui->show();

  int u0 = 20;
  int v0 = 50;
  int w = 648;
  int h = 484;
  fb = new FrameBuffer(u0, v0, w, h);
  fb->label("SW framebuffer");
  fb->show();
  gui->uiw->position(fb->w+20+20, 50);
//  Setup();
  ProjectiveTMSetup();
  Render();

}

void Scene::Render() {

  fb->FrameSetup(0xFF00AAFF);
  if (photo)
    fb->Set(photo);

  unsigned int colors[4] = {0xFF0000FF, 0xFF00FF00, 0xFFFF0000, 0xFFFFFFFF};
  for (int i = 0; i < markersN; i++) {
    fb->DrawPoint((int)markers[i][0], (int)markers[i][1], 13, colors[i]);
  }

  if (tm1)
//    tm1->RenderFilled(V3(1.0f, 0.0f, 0.0f), ppc, fb, 0, 0);
    tm1->RenderWireFrame(V3(1.0f, 0.0f, 0.0f), ppc, fb);
  if (tm2)
//    tm2->RenderFilled(V3(1.0f, 0.0f, 0.0f), ppc, fb, 0, 0);
    tm2->RenderWireFrame(V3(0.0f, 1.0f, 0.0f), ppc, fb);

  fb->redraw();

}

void Scene::ProjectiveTMSetup() {

  float hfov = 55.0f;
  ppc = new PPC(fb->w, fb->h, hfov);
  photo = new FrameBuffer(0, 0, 10, 10);
  photo->Load("mydbg/photoScene4.tif");

  tm1 = new TMesh();
  V3 center(2.5f, 0.0f, -5.0f);
  V3 xaxis(1.0f, 0.0f, 0.0f);
  V3 yaxis(0.0f, 0.0f, 1.0f);
  float rw = 5.0f;
  float rh = 10.0f;
  tm1->SetRectangle(center, xaxis, yaxis, rw, rh);
  tm1->colors = new V3[4];
  tm1->colors[0] = V3(1.0f, 0.0f, 0.0f);
  tm1->colors[1] = V3(0.0f, 1.0f, 0.0f);
  tm1->colors[2] = V3(0.0f, 0.0f, 1.0f);
  tm1->colors[3] = V3(1.0f, 1.0f, 1.0f);

  tm2 = new TMesh();
  V3 c0(-80.0f/12.0f, -29.0f/12.0f, -10.0f - 45.0f/12.0f);
  V3 c1(5.0f+(90.0f+80.0f)/12.0f, (29.0f + 79.0f)/12.0f, 10.0f + (45.0f + 69.0f)/12.0f);
  c1 = c1 + c0;
  AABB aabb;
  aabb.AddPoint(c0);
  aabb.AddPoint(c1);
  tm2->SetAABB(aabb);

}

void Scene::Setup() {

  float hfov = 55.0f;
  ppc = new PPC(fb->w, fb->h, hfov);
  tm1 = new TMesh();
  tm1->Load("geometry/teapot1k.bin");
  tm1->SetCenter(V3(0.0f, 0.0f, -100.0f));
  tm1->Scale(100.0f);

}

void Scene::DBG() {


  ppc->Load("mydbg/bestCamera.txt");
  Render();
  return;

//  ppc->Save("mydbg/photoCamera.txt");
  ppc->Load("mydbg/photoCamera.txt");
  Render();
  Fl::check();
  cerr << "INFO: initial guess error: " << ppc->ComputeError(tm1->verts, markers, 4) << " pix" << endl;
  ExtrinsicCalibrationBF();
  return;

  FrameBuffer *tmp = new FrameBuffer(0, 0, 10, 10);
  tmp->Load("mydbg/photoScene4.tif");
  tmp->label("loaded image");


//  tmp->DrawCircle(230.0f, 110.0f, 100.0f, 0xFF0000FF);

#if 0
  int u, v;
  for (v = 0; v < tmp->h/2; v++) {
    for (u = 0; u < tmp->w; u++) {
      unsigned int currColor = tmp->Get(u, v);
      unsigned int red = ((unsigned char*)&currColor)[0];
      red = 255-red;
      unsigned int green = ((unsigned char*)&currColor)[1];
      green = 255-green;
      unsigned int blue = ((unsigned char*)&currColor)[2];
      blue = 255-blue;
      currColor = 0xFF000000 + blue*256*256 + green*256 + red;
      tmp->Set(u, v, currColor);
    }
  }
#endif

  tmp->show();

  tmp->Save("mydbg/tmp.tif");


}


void Scene::ExtrinsicCalibrationBF() {

  // x, y, z, pan, tilt, roll
  int stepsN[6] = {15, 15, 15, 15, 15, 15};
  float amps[6] = {1.0f, 1.0f, 1.0f, 5.0f, 5.0f, 5.0f};
  int isteps[6];
  float minErr = FLT_MAX;
  float sol[6];
  PPC ppc0(*ppc);
  for (isteps[0] = 0; isteps[0] < stepsN[0]; isteps[0]++) {
    for (isteps[1] = 0; isteps[1] < stepsN[1]; isteps[1]++) {
      for (isteps[2] = 0; isteps[2] < stepsN[2]; isteps[2]++) {
        for (isteps[3] = 0; isteps[3] < stepsN[3]; isteps[3]++) {
          for (isteps[4] = 0; isteps[4] < stepsN[4]; isteps[4]++) {
            for (isteps[5] = 0; isteps[5] < stepsN[5]; isteps[5]++) {
              float currOffsets[6];
              for (int i = 0; i < 6; i++) {
                float frac = (float) isteps[i] / (float) (stepsN[i]-1);
                currOffsets[i] = -amps[i]/2.0f + amps[i]*frac;
              }
              PPC currPPC(*ppc);
              V3 currTrans(currOffsets[0], currOffsets[1], currOffsets[2]);
              currPPC.Translate(currTrans);
              currPPC.Pan(currOffsets[3]);
              currPPC.Tilt(currOffsets[4]);
              currPPC.Roll(currOffsets[5]);
              float currErr = currPPC.ComputeError(tm1->verts, markers, 4);
              if (currErr < minErr) {
                for (int i = 0; i < 6; i++) 
                  sol[i] = currOffsets[i];
                cerr << "INFO: found smaller error: " << currErr << "      \r";
                minErr = currErr;
                *ppc = currPPC;
                ppc->Save("mydbg/betterCamera.txt");
                Render();
                Fl::check();
                *ppc = ppc0;
              }
            }
          }
        }
      }
    }
  }
  cerr << endl << "INFO: DONE" << endl;

}
